#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netdb.h>
#include <assert.h>

#include "reliable_utils.h"

static int
do_recvpkt (int fd, char *buf, u_int pktlen,
	    struct sockaddr_in *from)
{
  socklen_t slen = sizeof (struct sockaddr_in);
  ssize_t rc;
  bzero (from, slen);

  // flag = 9
  rc = recvfrom (fd, buf, pktlen, 0, (struct sockaddr *) from, &slen);

  if (rc < 0 && (DEBUG_LVL & DEBUG_ERRORS))
    fprintf (stderr, "recvpkt: fd %d rc %d pktlen %d: %s [%d]\n", 
	     fd, rc, pktlen, strerror (errno), errno);
  else if (DEBUG_LVL & DEBUG_PKTV)
    fprintf (stderr, "recvpkt: fd %d rc %d pktlen %d\n", fd, rc, pktlen);

  if (rc > (int) pktlen) {
    if (DEBUG_LVL & DEBUG_ERRORS)
      fprintf (stderr, "recvpkt: max pkt len exceeded: %d > %d\n", rc, pktlen);
    rc = -1;
  }
  return rc;
}


static int
do_recvmsg (int fd, char *buf, u_int pktlen)
{
  //  char **lineptr;
  //  size_t *n;
  ssize_t rc = read (fd, buf, pktlen);

  if (rc > (int) pktlen) {
    if (DEBUG_LVL & DEBUG_ERRORS)
      fprintf (stderr, "recvmsg: max pkt len exceeded: %d > %d\n", rc, pktlen);
    rc = -1;
  }

  return rc;
}


int
sendpkt (int fd, struct sockaddr_in *to, char *buf, u_int len)
{
  ssize_t rc;

  assert (buf);
  assert (len);

  while (1) {
    rc = sendto (fd, buf, len, 0, 
		 (const struct sockaddr *) to, sizeof (struct sockaddr_in));

    if (!(rc < 0 && errno == EAGAIN)) {
      if (DEBUG_LVL & DEBUG_PKT)
	fprintf (stderr, "sendpkt [to %s:%d]: sent [%d/%d]: %s\n", 
		 inet_ntoa (to->sin_addr), ntohs (to->sin_port), 
		rc, len, strerror (errno));
      return rc;
    }
  }
}


void
recvpkt (int fd)
{
  char buf [pkt_bufsize];
  struct sockaddr_in from;
  socklen_t alen = sizeof (struct sockaddr_in);
  ssize_t rc;

  bzero (&buf, pkt_bufsize);
  bzero (&from, alen);

  if (fd <= 2)
    rc = do_recvmsg (fd, (char *) (&buf), pkt_bufsize);
  else 
    rc = do_recvpkt (fd, (char *) (&buf), pkt_mtu, &from);
  
  if (rc > 0) {
    handle_pkt (fd, (char *) (&buf), rc, from);
  }
  else if (!rc || (rc < 0 && errno != EAGAIN)) {
    if (DEBUG_LVL & DEBUG_ERRORS)
      fprintf (stderr, "recvpkt: rc [%d]: %s\n", rc, strerror (errno));
    rem_fd (fd);
  }
}
