#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "reliable_utils.h"

int want_timer;

static u_int32_t xid = 0;

// Timer delay is twice what timer_usec set to
#define timer_usec 50000

static void
timer_init ()
{
  if (!xid)
    xid = time (NULL) ^ (getpid () << 16);
}


static void
tvsub (struct timeval *a_minus_b, const struct timeval *a,
       const struct timeval *b)
{
  a_minus_b->tv_sec = a->tv_sec - b->tv_sec;
  if (a->tv_usec >= b->tv_usec)
    a_minus_b->tv_usec = a->tv_usec - b->tv_usec;
  else {
    a_minus_b->tv_usec = 1000000 + a->tv_usec - b->tv_usec;
    a_minus_b->tv_sec--;
  }
}

struct fd_t {
  int fd;
  LIST_ENTRY (fd_t) link;
};
static LIST_HEAD (fdlst_t, fd_t) fdlist = LIST_HEAD_INITIALIZER (fdlist);


void
add_fd (int fd)
{
  struct fd_t *fdt;
  if (!(fdt = malloc (sizeof (struct fd_t)))) {
    fprintf (stderr, "out of memory\n");
    exit (1);
  }

  fdt->fd = fd;
  LIST_INSERT_HEAD (&fdlist, fdt, link);
}

void
rem_fd (int fd)
{
  struct fd_t *fdt, *nfdt;
  for (fdt = LIST_FIRST (&fdlist); fdt; fdt = nfdt) {
    nfdt = LIST_NEXT (fdt, link);
    if (fdt->fd == fd) {
      LIST_REMOVE (fdt, link);
      free (fdt);
    }
  }
}


long
get_usec ()
{
  struct timespec tv;
  clock_gettime (CLOCK_REALTIME, &tv);
  return ((((long) (tv.tv_sec)) * 1000000) + (tv.tv_nsec / 1000));
}


void
run (void)
{
  fd_set rfds;
  struct timeval tvstart, tvnext, tvnow, tv, *tvp;
  struct fd_t *fdt, *nfdt;
  int n;

  timer_init ();
  gettimeofday (&tvstart, NULL);
  tvnow = tvnext = tvstart;

  for (;;) {
    FD_ZERO (&rfds);
    for (fdt = LIST_FIRST (&fdlist); fdt; fdt = nfdt) {
      nfdt = LIST_NEXT (fdt, link);
      FD_SET (fdt->fd, &rfds);      
    }

    tvp = NULL;
    if (want_timer > 0) {
      tvsub (&tv, &tvnext, &tvnow);
      tvp = &tv;
    }

    n = select (FD_SETSIZE, &rfds, NULL, NULL, tvp);
    for (fdt = LIST_FIRST (&fdlist); fdt; fdt = nfdt) {
      nfdt = LIST_NEXT (fdt, link);
      if (n > 0 && FD_ISSET (fdt->fd, &rfds)) {
	recvpkt (fdt->fd);
	FD_CLR (fdt->fd, &rfds);
	n--;
      }
    }

    gettimeofday (&tvnow, NULL);

    if (tvnow.tv_sec > tvnext.tv_sec
	|| (tvnow.tv_sec == tvnext.tv_sec && tvnow.tv_usec >= tvnext.tv_usec + timer_usec)) {

      int xtra = 1000000 - tvnext.tv_usec;
      if (xtra > timer_usec)
	tvnext.tv_usec = tvnow.tv_usec + timer_usec;
      else {
	tvnext.tv_sec++;
	tvnext.tv_usec = ((u_int) timer_usec - xtra);
      }

      if (want_timer > 0)
	timer ();
    }
  }
}

void 
setup_core ()
{
#if 0
  struct rlimit rlim;
  if (!getrlimit (RLIMIT_CORE, &rlim)) {
    rlim.rlim_cur = rlim.rlim_max;
    if (setrlimit (RLIMIT_CORE, &rlim))
      fprintf (stderr, "setrlimit (RLIMIT_CORE): %m");
    else if (!rlim.rlim_cur)
      fprintf (stderr,"cannot dump core");
    else {
      char buf[1024];
      char *wd = getcwd (buf, sizeof (buf));
      if (!wd)
	wd = "(no directory)";
      fprintf (stderr, "<= %d MB core files will go in %s\n",
	       ((int) (rlim.rlim_cur >> 20)), wd);
    }
  }
  else
    fprintf (stderr, "getrlimit (RLIMIT_CORE): %m");
#endif
}
